#if 0

Here is an example of how to use low level ( pre PB9 ) code to create an out of process COM server which can
be run as a simple stand alone exe by simply executing it, or it can be connected to through the COM system
in the manner of Excel or Word.  The program creates an object named class "CC", and the object exposes two
interfaces, e.g., an IX interface and an IY interface.  These interfaces contain the worthless functionality
of allowing an integer and a dynamic string to be 'Get/Set'. While useless in a practical sense, it does show 
how to create functionality in an out of process COM server object using low level PowerBASIC code.

If started as a stand alone exe by a user, it presents a visual interface where the IX and IY interfaces of
class CC can be exercised to store/retrieve integers and strings.  If started by COM's Service Control Manager
(after being registered by using the /r switch on the program's command line when started from a command prompt),
it won't provide any visual interface, but its COM interfaces will be available in the normal programatic manner.

My previous two COM examples ( CA and CB ) didn't have much in the way of instance variables, but this one CC -
does, i.e., the integer and string member variables.  The string is especially interesting in that it shows
how PowerBASIC strings must be dealt with low level, i.e., at the level of BSTRs, which are OLE COM length
prefixed null terminated strings.  To give you an idea of what I'm speaking of, if it hasn't already dawned on
you, if any of the interface functions must get/set a string, that string must be retained in a member variable
of the class CC - which in low level COM is created through a PowerBASIC Type.  And Types can't contain dynamic
variable length strings - only fixed length null terminated asciiz strings.  So, to store a BSTR (PowerBASIC
Dynamic String) in a Type one must revert to a more primitive form which would be some form of pointer variable
(I used Dword Ptrs here, but any kind of pointer would work, or even just a Dword.)....

Type CC                   'This is an example of what a class looks like in 'low level' COM
  lpIX                    As IXVtbl Ptr  'pointer to IX VTable
  lpIY                    As IYVtbl Ptr  'pointer to IY VTable
  m_iXInt                 As Long        'member (instance) variable of integral type in IX interface
  m_iYInt                 As Long        'member (instance) variable of integral type in IY interface
  m_XText                 As Dword Ptr   'BSTRs are allocated with the OLE String Engine using functions such as
  m_YText                 As Dword Ptr   'SysReAllocString() or SysAllocStringLen() and a pointer is returned.
  m_cRef                  As Long        'reference counting variables for outstanding interfaces of class
End Type

Check out the GetXText() / SetXText() and GetYText() / SetYText() interface functions and you'll see how that's
done.  Setting it up like this creates Strings that act exactly like BASIC Dynamic Strings to clients be they
Visual Basic or PowerBASIC.  This nastiness becomes an invisible and hidden implementation detail to clients.
To clients that use dynamic strings - they just look like bona fide strings (that's because they are bona fide
strings).

Another issue I want to touch on is 'marshalling'.  Marshalling in COM means the arrangement and movement of
interface function parameters and return values between processes.  There are three types of marshalling; custom
marshalling, standard marshalling, and type library or universal marshalling.  This program uses the easiest
which is universal or type library marshalling.

When this program, i.e., CC.Exe, is registered by starting it with the /r command line argument, the COM function
LoadTypeLibEx() is called in ExeRegisterServer(), and that function reads the TypeLib embedded in the Exe and
creates Interface keys for the IX and IY interfaces, and a typelib key under HKEY_CLASSES_ROOT.  Specifically, it 
creates all of the following keys in the registry under HKEY_CLASSES_ROOT:

1)  Creates HKEY_CLASSES_ROOT\TypeLib\{20000000-0000-0000-0000-000000000023}\1.0
2)  Creates HKEY_CLASSES_ROOT\TypeLib\{20000000-0000-0000-0000-000000000023}\1.0\0\win32=C:\Code\PwrBasic\PBWin90\CC\CC.exe '<Yours will be different!
3)  Creates HKEY_CLASSES_ROOT\TypeLib\{20000000-0000-0000-0000-000000000023}\Flags
4)  Creates HKEY_CLASSES_ROOT\TypeLib\{20000000-0000-0000-0000-000000000023}\1.0\HelpDir

5)  Creates HKEY_CLASSES_ROOT\Interface\$IID_IX
6)  Creates HKEY_CLASSES_ROOT\Interface\{20000000-0000-0000-0000-000000000021}\ProxyStubClsid={00020424-0000-0000-C000-000000000046}
7)  Creates HKEY_CLASSES_ROOT\Interface\{20000000-0000-0000-0000-000000000021}\ProxyStubClsid32={00020424-0000-0000-C000-000000000046}
8)  Creates HKEY_CLASSES_ROOT\Interface\{20000000-0000-0000-0000-000000000021}\TypeLib={20000000-0000-0000-0000-000000000023}

9)  Creates HKEY_CLASSES_ROOT\Interface\$IID_IY
10) Creates HKEY_CLASSES_ROOT\Interface\{20000000-0000-0000-0000-000000000022}\ProxyStubClsid={00020424-0000-0000-C000-000000000046}
11) Creates HKEY_CLASSES_ROOT\Interface\{20000000-0000-0000-0000-000000000022}\ProxyStubClsid32={00020424-0000-0000-C000-000000000046}
12) Creates HKEY_CLASSES_ROOT\Interface\{20000000-0000-0000-0000-000000000022}\TypeLib={20000000-0000-0000-0000-000000000023}

This is a good bit of stuff, to be sure.  None of this is necessary with in process dll servers, but due to the needs 
of inter-process marshalling, it becomes necessary for out of process servers.  These keys will be created if the 
oleautomation attribute is attached to the interface definitions in the interface definition file ( CC.idl ).  Here is
what the CC.idl file looks like...


//CC.idl
import "oaidl.idl";	

[object, uuid(20000000-0000-0000-0000-000000000021), oleautomation, helpstring("The IX Interface Functions")] //IX
interface IX : IUnknown
{
 HRESULT SetXInt([in] int iXVal);
 HRESULT GetXInt([out, retval] int* pInt);
 HRESULT SetXText([in] BSTR strText);
 HRESULT GetXText([out, retval] BSTR* strText);
};


[object, uuid(20000000-0000-0000-0000-000000000022), oleautomation, helpstring("The IY Interface Functions")] //IY
interface IY : IUnknown
{
 HRESULT SetYInt([in] int iYVal);
 HRESULT GetYInt([out, retval] int* pInt);
 HRESULT SetYText([in] BSTR strText);
 HRESULT GetYText([out, retval] BSTR* strText);
};

[uuid(20000000-0000-0000-0000-000000000023), version(1.0), helpstring("Class CC With TypeLib")] //LibId
library CCLibrary
{
 importlib("stdole32.tlb");
 [uuid(20000000-0000-0000-0000-000000000020)]
 coclass CC
 {
  [default] interface IX;
            interface IY;
 };
};
//End CC.idl


Note that due to the oleautomation attribute being attached to the interface definitions, the oleaut32.dll or universal
marshaller will be used to marshall method parameters.  If you locate... 

                       HKEY_CLASSES_ROOT\{00020424-0000-0000-C000-000000000046}\InProcServer32
                       
in your registry, you'll find its value set to oleaut32.dll, which is where the universal marshaller is located.  In practical 
terms to use the universal marshaller or Type Library Marshalling in your out of process servers you need to use only variant
compliant data types.  This includes the usual assembledge of various types of integers, floating point numbers, and BSTRs 
(dynamic basic strings), so its not to hard to live with.

I'll just say a word about standard marshalling.  When you use the Microsoft Interface Definition Language Compiler ( MIDL )
to compile a *.idl file into a Type Library, there are a number of additional *.c and *.h files produced as output besides the 
*.tlb file containing the Type Library.  Some of those other files can be compiled into proxy and stub code to be packaged with
a project, but how to do that is beyond the scope of what I wanted to cover here.  To do all that you would need to run a C/C++
compiler, and produce alternate data in the registry than listed above.

I'll try to include the CC.tlb file created by using the MIDL compiler on the above CC.idl file.  That way, if you are 
interested in compiling and running this example, but don't want to fool around with getting the MIDL compiler installed on
your system, you'll be able to do so.  The MIDL compiler is installed by many Microsoft products, including Visual Studio,
most or all of the Express versions of the same, as well as the Windows SDK.  So I'll try to provide the exact steps you need 
to take to create an out of process server using low level PowerBASIC code.  In the command line examples of mine below the
path to my project is...

C:\Code\PwrBasic\PBWin90\CC

So, yours may be different and just use yours instead of mine when you see see the above string.

First, compile the CC.rc file just below into a CC.pbr file using the PowerBASIC Windows compiler...


//CC.rc
 1  typelib CC.TLB
//End CC.rc 


Next, compile the CC.bas project in your PowerBASIC Windows compiler which includes the CC.inc, Main.inc, and Registry.inc
files.  Note that you need to use Jose's includes!  After you compile the CC.bas file you'll end up with CC.exe, and you can 
run that if you like and see the functionality of the program through the user interface elements.  However, at this point
it is running as a stand alone exe and the COM subsystem of Windows knows nothing about it and won't be able to connect
to it as an out of process server.  For Windows to know about it it must be registered, and to register it we need/want
to embed the type library for the COM Class CC in the exe file.

Take the CC.idl file listed above or the one included in the zip file and compile it with MIDL like so...


C:\Code\PwrBasic\PBWin90\CC>midl cc.idl 
 

After successfully compiling that there should be a CC.tlb file in your working directory.  Next use the PowerBASIC TypeLib
Embedder program PBTyp.exe to embed the type library in the CC.exe file just produced by PowerBASIC Windows...


C:\Code\PwrBasic\PBCC50\CC>PBTyp CC.Exe CC.rc


Here is the output from that on my machine from the PBTyp.exe program in your PowerBASIC installation...


PowerBASIC PBTYP TypeLib Embedder - Rev 1.0
Copyright (c) 2007 PowerBASIC Inc.

Module:   CC
Target:   CC.EXE
TypeLib:  CC.TLB
Resource: CC.RC

That should be it and you are ready to register everything with Windows!  At the command prompt window where you have
the CC.exe file execute this at the command line...


C:\Code\PwrBasic\PBWin90\CC>CC.exe /r

On my XP or Win2000 systems I get output like the following (cleaned up a bit) after that step in an AllocConsole()
window...

Entering WinMain()
  lpCmdLine = /r
  Entering Initialize()
    IClassFactory_Vtbl.QueryInterface =  4219397
    IClassFactory_Vtbl.AddRef         =  4219677
    IClassFactory_Vtbl.Release        =  4219816
    IClassFactory_Vtbl.CreateInstance =  4219963
    IClassFactory_Vtbl.LockServer     =  4221600
    Varptr(CCClassFactory)            =  4257784
    Varptr(CCClassFactory.lpVtbl)     =  4257784
    Varptr(IClassFactory_Vtbl)        =  4257788
    CCClassFactory.lpVtbl             =  4257788
    Called CCClassFactory_QueryInterface()
      Called CCClassFactory_AddRef()!
    Leaving CCClassFactory_QueryInterface()
    pClsFac =  4257784
  Leaving Initialize()

Entering blnCmdLineProcessing()
  Calling ExeRegisterServer()
  Entering ExeRegisterServer()
    szPath         = C:\Code\PwrBasic\PBWin90\CC\CC.EXE
    LoadTypeLib() Succeeded!
    Entering RegisterServer()
      szExeName  = C:\Code\PwrBasic\PBWin90\CC\CC.EXE
      szClsid    = {20000000-0000-0000-0000-000000000020}
      szLibid    = {20000000-0000-0000-0000-000000000023}
      szKey      = CLSID\{20000000-0000-0000-0000-000000000020}
  Leaving ExeRegisterServer()
  ExeRegisterServer() Apparently Succeeded!
Leaving blnCmdLineProcessing()

At this point you'll probably want to examine your registry under HKEY_CLASSES_ROOT in terms of CLSID, Program ID,
Interface, and TypeLib keys to see everything I've been discussing to this point.  After this post I'll try to 
provide various example clients in C, C++, VB6, VB.NET, and PowerBASIC to exercise the local exe server. Finally,
you may find additional information about this project from when I first developed it several months ago at this
link including Console Compiler versions...

http://www.jose.it-berater.org/smfforum/index.php?topic=3666.0 

#endif

'Tested With PowerBASIC Windows Compiler 9.04 And Jose Roca's Includes v116
#Compile Exe              "CC.Exe"            'C:\Code\PwrBasic\PBWin90\CC\CC.Exe
#Dim                      All                 'This program requires Jose Roca's includes.  Jose's includes must
#Register                 None                'not be intermixed with the PowerBASIC includes obtained with
#Include                  "Win32Api.inc"      'purchase of the PowerBASIC compiler and installed in the WinApi
#Include                  "ObjBase.inc"       'directory under the PowerBASIC installation path.  Jose's includes
#Include                  "OAIdl.inc"         'should be saved to an addition empty directory such as WinApiEx or
#Include                  "Main.inc"          'something like that.  Under the Options menu of PowerBASIC you can
#Include                  "Registry.inc"      'then Set the Include Path to Jose's Includes.
#Include                  "CC.inc"
#Resource                 "CC.pbr"

Function fnWndProc_OnCreate(wea As WndEventArgs) As Long        'This function will execute whether this program is
  Local pCreateStruct As CREATESTRUCT Ptr                       'started by a user in the manner of a standard exe,
  Local lpCmdLine As Asciiz Ptr                                 'or whether it was started by the COM Service Control
  Local hCtl,pUnk As Dword                                      'Manager (SCM).  If this program was started by SCM
  Local Vtbl As Dword Ptr                                       'it won't create any visible window (it will create
  Local hr As Long                                              'a non-visible window though).

  Prnt "  Entering fnWndProc_OnCreate()", 1                     'Here we want to check if we were loaded by COM's
  pCreateStruct=wea.lParam                                      'Service Control Manager (SCM).  If we were there
  wea.hInst=@pCreateStruct.hInstance                            'would be a "/Embedding" or "-Embedding" in the
  lpCmdLine=@pCreateStruct.lpCreateParams                       'lpCmdLine parameter, i.e., its length wouldn't be
  Prnt "    lpCmdLine       = " & Str$(lpCmdLine), 1            'zero.  The logic just left and below checks this
  Prnt "    @lpCmdLine      = " & @lpCmdLine, 1                 'and if the length of this string is zero then this
  Prnt "    Len(@lpCmdLine) = " & Str$(Len(@lpCmdLine)), 1      'program is to run as a stand alone Exe and present
  If Len(@lpCmdLine)=0 Then
     hr=CCClassFactory_CreateInstance(Varptr(CCClassFactory), pUnk, $IID_IX, Varptr(pIX))
     Prnt "    pIX          = " & Str$(pIX), 1
     Prnt "    @pIX         = " & Str$(@pIX), 1                 'a user interface to the user.  Therefore it creates
     Vtbl=@pIX                                                  'buttons, labels, and text boxes, and executes a
     Prnt "    Vtbl         = " & Str$(Vtbl), 1                 'ShowWindow() call at bottom within the enclosing If.
     If FAILED(hr) Then                                         'If the program was started by COM, then no
        Function=-1 : Exit Function                             'ShowWindow() is executed, but the window is non-
     End If                                                     'theless created not visible.
     Call Dword @Vtbl[0] Using ptrQueryInterface(pIX, $IID_IY, pIY) To hr
     If SUCCEEDED(hr) Then
        Prnt "    pIX->QueryInterface(pIY) Succeeded!", 1       'All these CreateWindowEx() calls below are important if this
     Else                                                       'program was started as a stand alone exe by a client (not by
        Prnt "    pIX->QueryInterface(pIY) Failed!", 1          'COM's SCM), as they will provide user interface elements
     End If                                                     'with which the user can interact on the visible main program
     Prnt "    pIY  = " & Str$(pIY), 1                          'window.
     Prnt "    @pIY = " & Str$(@pIY), 1
     hCtl=CreateWindowEx(0,"static","The IX Interface",%WS_CHILD Or %WS_VISIBLE,215,5,120,20,Wea.hWnd,-1,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(0,"static","Set X Int",%WS_CHILD Or %WS_VISIBLE,10,35,80,20,Wea.hWnd,-1,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(%WS_EX_CLIENTEDGE,"edit","",%WS_CHILD Or %WS_VISIBLE,100,35,40,22,Wea.hWnd,%EDIT_SET_X_INT,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(0,"button","Set X Int",%WS_CHILD Or %WS_VISIBLE,160,35,75,22,Wea.hWnd,%BTN_SET_X_INT,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(0,"static","Get X Int",%WS_CHILD Or %WS_VISIBLE,280,35,80,20,Wea.hWnd,-1,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(%WS_EX_CLIENTEDGE,"edit","",%WS_CHILD Or %WS_VISIBLE,380,35,40,22,Wea.hWnd,%EDIT_GET_X_INT,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(0,"button","Get X Int",%WS_CHILD Or %WS_VISIBLE,440,35,75,22,Wea.hWnd,%BTN_GET_X_INT,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(0,"static","Set X Text",%WS_CHILD Or %WS_VISIBLE,10,75,80,20,Wea.hWnd,-1,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(%WS_EX_CLIENTEDGE,"edit","",%WS_CHILD Or %WS_VISIBLE,100,75,320,22,Wea.hWnd,%EDIT_SET_X_TEXT,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(0,"button","Set X Text",%WS_CHILD Or %WS_VISIBLE,440,75,75,22,Wea.hWnd,%BTN_SET_X_TEXT,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(0,"static","Get X Text",%WS_CHILD Or %WS_VISIBLE,10,115,110,20,Wea.hWnd,-1,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(%WS_EX_CLIENTEDGE,"edit","",%WS_CHILD Or %WS_VISIBLE,100,115,320,22,Wea.hWnd,%EDIT_GET_X_TEXT,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(0,"button","Get X Text",%WS_CHILD Or %WS_VISIBLE,440,115,75,22,Wea.hWnd,%BTN_GET_X_TEXT,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(0,"static","The IY Interface",%WS_CHILD Or %WS_VISIBLE,215,170,120,20,Wea.hWnd,-1,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(0,"static","Set Y Int",%WS_CHILD Or %WS_VISIBLE,10,195,80,20,Wea.hWnd,-1,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(%WS_EX_CLIENTEDGE,"edit","",%WS_CHILD Or %WS_VISIBLE,100,195,40,22,Wea.hWnd,%EDIT_SET_Y_INT,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(0,"button","Set Y Int",%WS_CHILD Or %WS_VISIBLE,160,195,75,22,Wea.hWnd,%BTN_SET_Y_INT,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(0,"static","Get Y Int",%WS_CHILD Or %WS_VISIBLE,280,195,80,20,Wea.hWnd,-1,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(%WS_EX_CLIENTEDGE,"edit","",%WS_CHILD Or %WS_VISIBLE,380,195,40,22,Wea.hWnd,%EDIT_GET_Y_INT,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(0,"button","Get Y Int",%WS_CHILD Or %WS_VISIBLE,440,195,75,22,Wea.hWnd,%BTN_GET_Y_INT,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(0,"static","Set Y Text",%WS_CHILD Or %WS_VISIBLE,10,235,80,20,Wea.hWnd,-1,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(%WS_EX_CLIENTEDGE,"edit","",%WS_CHILD Or %WS_VISIBLE,100,235,320,22,Wea.hWnd,%EDIT_SET_Y_TEXT,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(0,"button","Set Y Text",%WS_CHILD Or %WS_VISIBLE,440,235,75,22,Wea.hWnd,%BTN_SET_Y_TEXT,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(0,"static","Get Y Text",%WS_CHILD Or %WS_VISIBLE,10,265,110,20,Wea.hWnd,-1,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(%WS_EX_CLIENTEDGE,"edit","",%WS_CHILD Or %WS_VISIBLE,100,265,320,22,Wea.hWnd,%EDIT_GET_Y_TEXT,Wea.hInst,Byval 0)
     hCtl=CreateWindowEx(0,"button","Get Y Text",%WS_CHILD Or %WS_VISIBLE,440,265,75,22,Wea.hWnd,%BTN_GET_Y_TEXT,Wea.hInst,Byval 0)
     Call ShowWindow(Wea.hWnd, %SW_SHOWNORMAL)
  End If
  Prnt "  Leaving fnWndProc_OnCreate()", 1

  fnWndProc_OnCreate=0
End Function


Function fnWndProc_OnCommand(Wea As WndEventArgs) As Long
  Local szBuffer As Asciiz*128
  Local strBuffer As String
  Local Vtbl As Dword Ptr
  Local x,y As Long

  Vtbl=@pIX
  Select Case As Long Lowrd(Wea.wParam)
    Case %BTN_SET_X_INT
      Call GetWindowText(GetDlgItem(Wea.hWnd,%EDIT_SET_X_INT),szBuffer,16)
      x=Val(szBuffer)
      Call Dword @Vtbl[3] Using ptrSetInt(pIX,x)
    Case %BTN_GET_X_INT
      Call Dword @Vtbl[4] Using ptrGetInt(pIX,x)
      szBuffer=Str$(x)
      Call SetWindowText(GetDlgItem(Wea.hWnd,%EDIT_GET_X_INT),szBuffer)
    Case %BTN_SET_X_TEXT
      Call GetWindowText(GetDlgItem(Wea.hWnd,%EDIT_SET_X_TEXT),szBuffer,128)
      strBuffer=szBuffer
      strBuffer=UCode$(strBuffer)
      Call Dword @Vtbl[5] Using ptrSetText(pIX, Byval strBuffer)
    Case %BTN_GET_X_TEXT
      Call Dword @Vtbl[6] Using ptrGetText(pIX, Byref strBuffer)
      strBuffer=ACode$(strBuffer)
      Call SetWindowText(GetDlgItem(Wea.hWnd,%EDIT_GET_X_TEXT),Byval Strptr(strBuffer))
    Case %BTN_SET_Y_INT
      Call GetWindowText(GetDlgItem(Wea.hWnd,%EDIT_SET_Y_INT),szBuffer,16)
      y=Val(szBuffer)
      Call Dword @Vtbl[3] Using ptrSetInt(pIY,y)
    Case %BTN_GET_Y_INT
      Call Dword @Vtbl[4] Using ptrGetInt(pIY,y)
      szBuffer=Str$(y)
      Call SetWindowText(GetDlgItem(Wea.hWnd,%EDIT_GET_Y_INT),szBuffer)
    Case %BTN_SET_Y_TEXT
      Call GetWindowText(GetDlgItem(Wea.hWnd,%EDIT_SET_Y_TEXT),szBuffer,128)
      strBuffer=szBuffer
      strBuffer=UCode$(strBuffer)
      Call Dword @Vtbl[5] Using ptrSetText(pIY, Byval strBuffer)
    Case %BTN_GET_Y_TEXT
      Call Dword @Vtbl[6] Using ptrGetText(pIY, Byref strBuffer)
      strBuffer=ACode$(strBuffer)
      Call SetWindowText(GetDlgItem(Wea.hWnd,%EDIT_GET_Y_TEXT),Byval Strptr(strBuffer))
  End Select

  fnWndProc_OnCommand=0
End Function


Function fnWndProc_OnPaint(Wea As WndEventArgs) As Long
  Local ps As PAINTSTRUCT
  Local hDC As Dword

  hDC=BeginPaint(Wea.hWnd, ps)
  MoveToEx(hDC, 20, 155, Byval 0)
  LineTo(hDC, 510, 155)
  EndPaint(Wea.hWnd, ps)

  fnWndProc_OnPaint=0
End Function


Function fnWndProc_OnClose(Wea As WndEventArgs) As Long
  Local Vtbl As Dword Ptr
  Local hr As Long
                                                    'Control flow can reach this procedure in either
  Prnt "  Entering fnWndProc_OnClose()", 1          'of two ways.  If the program is running as a
  Call DestroyWindow(Wea.hWnd)                      'stand alone executable execution will occur here
  Prnt "    pIX  = " & Str$(pIX), 1                 'when the user clicks the 'x' button in the Window's
  Prnt "    @pIX = " & Str$(@pIY), 1                'title bar.  Alternately, if the program was loaded
  Vtbl=@pIX                                         'by COM, when the client releases its interface
  Call DWord @VTbl[2] Using ptrRelease(pIX) To hr   'pointers the reference count held in the g_lLocks
  Prnt "    pIY  = " & Str$(pIY), 1                 'global will reach zero in UnLock(), and a WM_CLOSE
  Prnt "    @pIY = " & Str$(@pIY), 1                'message will be sent to this window, as well as a
  Vtbl=@pIY                                         'PostQuitMessage() to terminate the message queene.
  Call DWord @VTbl[2] Using ptrRelease(pIY) To hr
  Prnt "  Leaving fnWndProc_OnClose()", 1

  fnWndProc_OnClose=0
End Function


Function fnWndProc(ByVal hWnd As Long,ByVal wMsg As Long,ByVal wParam As Long,ByVal lParam As Long) As Long
  Static wea As WndEventArgs
  Register iReturn As Long
  Register i As Long

  For i=0 To 3
    If wMsg=MsgHdlr(i).wMessage Then
       wea.hWnd=hWnd: wea.wParam=wParam: wea.lParam=lParam
       Call Dword MsgHdlr(i).dwFnPtr Using FnPtr(wea) To iReturn
       fnWndProc=iReturn
       Exit Function
    End If
  Next i

  fnWndProc=DefWindowProc(hWnd,wMsg,wParam,lParam)
End Function


Sub AttachMessageHandlers()
  ReDim MsgHdlr(3) As MessageHandler  'Associate Windows Message With Message Handlers
  MsgHdlr(0).wMessage=%WM_CREATE   :   MsgHdlr(0).dwFnPtr=CodePtr(fnWndProc_OnCreate)
  MsgHdlr(1).wMessage=%WM_COMMAND  :   MsgHdlr(1).dwFnPtr=CodePtr(fnWndProc_OnCommand)
  MsgHdlr(2).wMessage=%WM_PAINT    :   MsgHdlr(2).dwFnPtr=CodePtr(fnWndProc_OnPaint)
  MsgHdlr(3).wMessage=%WM_CLOSE    :   MsgHdlr(3).dwFnPtr=CodePtr(fnWndProc_OnClose)
End Sub


Sub Terminate(Byval lpCmdLine As Asciiz Ptr, Byref regID As Dword)
  If InStr(@lpCmdLine,"/Embedding") Or InStr(@lpCmdLine,"-Embedding") Then
     Call CoRevokeClassObject(regID)
  End If
End Sub


Function WinMain(ByVal hInstance As Long, ByVal hPrev As Long, ByVal lpCmdLine As Asciiz Ptr, ByVal iShow As Long) As Long
  Local szAppName As Asciiz*8
  Local blnFailure As Long
  Local wc As WndClassEx
  Local regID As Dword
  Local Msg As tagMsg

  Call AllocConsole()                      'Create console for debug output.  If this program was started by COM as a result,
  Prnt "Entering WinMain()", 1             'for example, of a CoCreateInstance() call or high level call in PowerBASIC such as
  Prnt "  lpCmdLine = " & @lpCmdLine, 1    'Let pIX = AnyCom "ComObject.CC", then lpCmdLine will point to a string containing
  Call Initialize()                        '-Embedding or /Embedding.  In any case, subs Initialize() and AttachEventHandlers()
  Call AttachMessageHandlers()             'have to be called.  Initialize() creates an IClassFactory1 object which is an imple-
  If CmdLineProcessing(hInstance, lpCmdLine, regID, blnFailure) Then
     Waitkey()                             'mentation of an important COM concept know as a Class Factory.  Class Factories are
     Function=0 :  Exit Function           'something like the 'new' operator in C++ in that they create a specific kind of
  End If                                   'object.  See CCClassFactory_CreateInstance() in CC.inc.  That is where a CC object
  If blnFailure Then                       'is created with low level COM code.
     MessageBox(%NULL,"CoRegisterClassObject() Failed!  This Is Decidedly Bad!","Error Report!",%MB_ICONERROR)
     Function=-1 : Exit Function
  End If
  szAppName="CC.Exe"
  wc.cbSize=SizeOf(wc)                              :   wc.style=%CS_HREDRAW Or %CS_VREDRAW
  wc.lpfnWndProc=CodePtr(fnWndProc)                 :   wc.cbClsExtra=0
  wc.cbWndExtra=0                                   :   wc.hInstance=hInstance
  wc.hIcon=LoadIcon(%NULL, ByVal %IDI_APPLICATION)  :   wc.hCursor=LoadCursor(%NULL, ByVal %IDC_ARROW)
  wc.hbrBackground=%COLOR_BTNFACE+1                 :   wc.lpszMenuName=%NULL
  wc.lpszClassName=VarPtr(szAppName)
  Call RegisterClassEx(wc)
  hMainWnd=CreateWindowEx(0, szAppName, szAppName, %WS_OVERLAPPEDWINDOW, 400, 200, 545, 350, 0, 0, hInstance, ByVal lpCmdLine)
  While GetMessage(Msg,%NULL,0,0)
    TranslateMessage Msg
    DispatchMessage Msg
  Wend
  Call Terminate(lpCmdLine, regID)
  Prnt "Leaving WinMain()", 1
  Waitkey()

  WinMain=0
End Function
